//! Stores a [Token's](super::Token) lexical information
use std::{fmt::Display, str::FromStr};

/// Stores a [Token's](super::Token) lexical information
#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
pub enum TokenKind {
    /// Invalid sequence
    Invalid,
    /// Any kind of comment
    Comment,
    /// Any tokenizable literal (See [TokenData](super::TokenData))
    Literal,
    /// A non-keyword identifier
    Identifier,
    // A keyword
    As,       // as
    Break,    // "break"
    Cl,       // "cl"
    Const,    // "const"
    Continue, // "continue"
    Else,     // "else"
    Enum,     // "enum"
    False,    // "false"
    Fn,       // "fn"
    For,      // "for"
    If,       // "if"
    Impl,     // "impl"
    In,       // "in"
    Let,      // "let"
    Loop,     // "loop"
    Match,    // "match"
    Mod,      // "mod"
    Mut,      // "mut"
    Pub,      // "pub"
    Return,   // "return"
    SelfKw,   // "self"
    SelfTy,   // "Self"
    Static,   // "static"
    Struct,   // "struct"
    Super,    // "super"
    True,     // "true"
    Type,     // "type"
    Use,      // "use"
    While,    // "while"
    // Delimiter or punctuation
    LCurly,     // {
    RCurly,     // }
    LBrack,     // [
    RBrack,     // ]
    LParen,     // (
    RParen,     // )
    Amp,        // &
    AmpAmp,     // &&
    AmpEq,      // &=
    Arrow,      // ->
    At,         // @
    Backslash,  // \
    Bang,       // !
    BangBang,   // !!
    BangEq,     // !=
    Bar,        // |
    BarBar,     // ||
    BarEq,      // |=
    Colon,      // :
    ColonColon, // ::
    Comma,      // ,
    Dot,        // .
    DotDot,     // ..
    DotDotEq,   // ..=
    Eq,         // =
    EqEq,       // ==
    FatArrow,   // =>
    Grave,      // `
    Gt,         // >
    GtEq,       // >=
    GtGt,       // >>
    GtGtEq,     // >>=
    Hash,       // #
    HashBang,   // #!
    Lt,         // <
    LtEq,       // <=
    LtLt,       // <<
    LtLtEq,     // <<=
    Minus,      // -
    MinusEq,    // -=
    Plus,       // +
    PlusEq,     // +=
    Question,   // ?
    Rem,        // %
    RemEq,      // %=
    Semi,       // ;
    Slash,      // /
    SlashEq,    // /=
    Star,       // *
    StarEq,     // *=
    Tilde,      // ~
    Xor,        // ^
    XorEq,      // ^=
    XorXor,     // ^^
}

impl Display for TokenKind {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            TokenKind::Invalid => "invalid".fmt(f),
            TokenKind::Comment => "comment".fmt(f),
            TokenKind::Literal => "literal".fmt(f),
            TokenKind::Identifier => "identifier".fmt(f),

            TokenKind::As => "sama".fmt(f),
            TokenKind::Break => "pana".fmt(f),
            TokenKind::Cl => "la".fmt(f),
            TokenKind::Const => "kiwen".fmt(f),
            TokenKind::Continue => "tawa".fmt(f),
            TokenKind::Else => "taso".fmt(f),
            TokenKind::Enum => "kulupu".fmt(f),
            TokenKind::False => "ike".fmt(f),
            TokenKind::Fn => "nasin".fmt(f),
            TokenKind::For => "ale".fmt(f),
            TokenKind::If => "tan".fmt(f),
            TokenKind::Impl => "insa".fmt(f),
            TokenKind::In => "lon".fmt(f),
            TokenKind::Let => "poki".fmt(f),
            TokenKind::Loop => "awen".fmt(f),
            TokenKind::Match => "seme".fmt(f),
            TokenKind::Mod => "selo".fmt(f),
            TokenKind::Mut => "ante".fmt(f),
            TokenKind::Pub => "lukin".fmt(f),
            TokenKind::Return => "pini".fmt(f),
            TokenKind::SelfKw => "mi".fmt(f),
            TokenKind::SelfTy => "Mi".fmt(f),
            TokenKind::Static => "mute".fmt(f),
            TokenKind::Struct => "lipu".fmt(f),
            TokenKind::Super => "mama".fmt(f),
            TokenKind::True => "pona".fmt(f),
            TokenKind::Type => "ijo".fmt(f),
            TokenKind::Use => "jo".fmt(f),
            TokenKind::While => "lawa".fmt(f),

            TokenKind::LCurly => "{".fmt(f),
            TokenKind::RCurly => "}".fmt(f),
            TokenKind::LBrack => "[".fmt(f),
            TokenKind::RBrack => "]".fmt(f),
            TokenKind::LParen => "(".fmt(f),
            TokenKind::RParen => ")".fmt(f),
            TokenKind::Amp => "&".fmt(f),
            TokenKind::AmpAmp => "&&".fmt(f),
            TokenKind::AmpEq => "&=".fmt(f),
            TokenKind::Arrow => "->".fmt(f),
            TokenKind::At => "@".fmt(f),
            TokenKind::Backslash => "\\".fmt(f),
            TokenKind::Bang => "!".fmt(f),
            TokenKind::BangBang => "!!".fmt(f),
            TokenKind::BangEq => "!=".fmt(f),
            TokenKind::Bar => "|".fmt(f),
            TokenKind::BarBar => "||".fmt(f),
            TokenKind::BarEq => "|=".fmt(f),
            TokenKind::Colon => ":".fmt(f),
            TokenKind::ColonColon => "::".fmt(f),
            TokenKind::Comma => ",".fmt(f),
            TokenKind::Dot => ".".fmt(f),
            TokenKind::DotDot => "..".fmt(f),
            TokenKind::DotDotEq => "..=".fmt(f),
            TokenKind::Eq => "=".fmt(f),
            TokenKind::EqEq => "==".fmt(f),
            TokenKind::FatArrow => "=>".fmt(f),
            TokenKind::Grave => "`".fmt(f),
            TokenKind::Gt => ">".fmt(f),
            TokenKind::GtEq => ">=".fmt(f),
            TokenKind::GtGt => ">>".fmt(f),
            TokenKind::GtGtEq => ">>=".fmt(f),
            TokenKind::Hash => "#".fmt(f),
            TokenKind::HashBang => "#!".fmt(f),
            TokenKind::Lt => "<".fmt(f),
            TokenKind::LtEq => "<=".fmt(f),
            TokenKind::LtLt => "<<".fmt(f),
            TokenKind::LtLtEq => "<<=".fmt(f),
            TokenKind::Minus => "-".fmt(f),
            TokenKind::MinusEq => "-=".fmt(f),
            TokenKind::Plus => "+".fmt(f),
            TokenKind::PlusEq => "+=".fmt(f),
            TokenKind::Question => "?".fmt(f),
            TokenKind::Rem => "%".fmt(f),
            TokenKind::RemEq => "%=".fmt(f),
            TokenKind::Semi => ";".fmt(f),
            TokenKind::Slash => "/".fmt(f),
            TokenKind::SlashEq => "/=".fmt(f),
            TokenKind::Star => "*".fmt(f),
            TokenKind::StarEq => "*=".fmt(f),
            TokenKind::Tilde => "~".fmt(f),
            TokenKind::Xor => "^".fmt(f),
            TokenKind::XorEq => "^=".fmt(f),
            TokenKind::XorXor => "^^".fmt(f),
        }
    }
}
impl FromStr for TokenKind {
    /// [FromStr] can only fail when an identifier isn't a keyword
    type Err = ();
    /// Parses a string s to return a Keyword
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(match s {
            "as" | "sama" => Self::As,
            "break" | "pana" => Self::Break,
            "cl" | "la" => Self::Cl,
            "const" | "kiwen" => Self::Const,
            "continue" | "tawa" => Self::Continue,
            "else" | "taso" => Self::Else,
            "enum" | "kulupu" => Self::Enum,
            "false" | "ike" => Self::False,
            "fn" | "nasin" => Self::Fn,
            "for" | "ale" => Self::For,
            "if" | "tan" => Self::If,
            "impl" | "insa" => Self::Impl,
            "in" | "lon" => Self::In,
            "let" | "poki" => Self::Let,
            "loop" | "awen" => Self::Loop,
            "match" | "seme" => Self::Match,
            "mod" | "selo" => Self::Mod,
            "mut" | "ante" => Self::Mut,
            "pub" | "lukin" => Self::Pub,
            "return" | "pini" => Self::Return,
            "self" | "mi" => Self::SelfKw,
            "Self" | "Mi" => Self::SelfTy,
            "static" | "mute" => Self::Static,
            "struct" | "lipu" => Self::Struct,
            "super" | "mama" => Self::Super,
            "true" | "pona" => Self::True,
            "type" | "ijo" => Self::Type,
            "use" | "jo" => Self::Use,
            "while" | "lawa" => Self::While,
            _ => Err(())?,
        })
    }
}
